/*
  File name: I3Dg.cpp
  Copyright: (C) 2003 - 2007 Tomasz Kazmierczak

  Creation date: 20.10.2003
  Last modification date: 23.09.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#include <math.h>
#include "I3Dg.h"

I3Dg *create3DgInterface()
{
	return new I3Dg;
}

I3Dg::I3Dg() :
	mError(ERR3DG_OK), mLightsQuantity(0), mTexturesQuantity(0),
	mMaterialsQuantity(0), mLights(0), mTextures(0), mMaterials(0)
{
}

I3Dg::~I3Dg()
{
	if(mTextures)
	{
		for(uint32_t i = 0; i < mTexturesQuantity; i++)
		{
			if(mTextures[i])
			{
				if(mTextures[i]->pPath)
					delete[] mTextures[i]->pPath;
				delete mTextures[i];
			}
		}
		delete[] mTextures;
	}
	if(mMaterials)
	{
		for(uint32_t i = 0; i < mMaterialsQuantity; i++)
		{
			if(mMaterials[i])
				delete mMaterials[i];
		}
		delete[] mMaterials;
	}
	while(not mMeshes.empty())
	{
		if(mMeshes.begin()->second)
			delete mMeshes.begin()->second;
		mMeshes.erase(mMeshes.begin());
	}
	while(not mMeshBases.empty())
	{
		if(mMeshBases.begin()->second)
			delete mMeshBases.begin()->second;
		mMeshBases.erase(mMeshBases.begin());
	}
	if(mLights)
	{
		for(uint32_t i = 0; i < mLightsQuantity; i++)
		{
			if(mLights[i])
				delete mLights[i];
		}
		delete[] mLights;
	}
}

void I3Dg::release()
{
	delete this;
}

MeshBase3DgUID I3Dg::generateMeshBaseUID()
{
	if(mMeshBases.empty())
		return 0;
	MeshBaseMap3Dg::iterator i = mMeshBases.end();
	i--;
	MeshBase3DgUID uid = i->second->getUID();
	//uid is a 32 bit data type, so this loop is quite safe
	do
	{
		uid++;
		i = mMeshBases.find(uid);
	}
	while(i != mMeshBases.end());
	return uid;
}

Mesh3DgUID I3Dg::generateMeshUID()
{
	if(mMeshes.empty())
		return 0;
	MeshMap3Dg::iterator i = mMeshes.end();
	i--;
	Mesh3DgUID uid = i->second->getUID();
	//uid is a 32 bit data type, so this loop is quite safe
	do
	{
		uid++;
		i = mMeshes.find(uid);
	}
	while(i != mMeshes.end());
	return uid;
}

MeshBase3Dg *I3Dg::createMeshBase(MeshBase3DgUID uiMBUID)
{
	//first check if there is already a base with this uid
	MeshBaseMap3Dg::iterator i = mMeshBases.find(uiMBUID);
	//if there is
	if(i != mMeshBases.end())
	{
		mError = ERR3DG_INVALIDARG;
		return 0;
	}

	MeshBase3Dg *pBase = new MeshBase3Dg(uiMBUID);
	if(not pBase)
	{
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}
	pBase->setParent(this);
	mMeshBases[uiMBUID] = pBase;
	return pBase;
}

Mesh3Dg *I3Dg::createMesh(Mesh3DgUID uiMUID)
{
	//first check if there is already a mesh with this uid
	MeshMap3Dg::iterator i = mMeshes.find(uiMUID);
	//if there is
	if(i != mMeshes.end())
	{
		mError = ERR3DG_INVALIDARG;
		return 0;
	}

	Mesh3Dg *pMesh = new Mesh3Dg(uiMUID);
	if(not pMesh)
	{
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}
	pMesh->setParent(this);
	mMeshes[uiMUID] = pMesh;
	return pMesh;
}

void I3Dg::destroyMesh(Mesh3DgUID uiMUID)
{
	//find the mesh
	MeshMap3Dg::iterator i = mMeshes.find(uiMUID);
	//if no such mesh
	if(i == mMeshes.end())
		return;

	//remove the mesh from memory
	delete i->second;
	//erase the map entry
	mMeshes.erase(i);
}

Light3Dg *I3Dg::createLight()
{
	Light3Dg **pTempLights = new Light3Dg*[mLightsQuantity+1];
	if(not pTempLights)
	{
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	Light3Dg *pLight = new Light3Dg;
	//FIXME: when light won't be a simple struct it will have to be removed
	memset(pLight, 0, sizeof(Light3Dg));
	if(not pLight)
	{
		delete[] pTempLights;
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	for(uint32_t i = 0; i < mLightsQuantity; i++)
		pTempLights[i] = mLights[i];

	delete[] mLights;
	mLights = pTempLights;
	mLights[mLightsQuantity] = pLight;
	mLightsQuantity++;
	return pLight;
}

Material3Dg *I3Dg::createMaterial()
{
	Material3Dg **pTempMaterials = new Material3Dg*[mMaterialsQuantity+1];
	if(not pTempMaterials)
	{
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	Material3Dg *pMaterial = new Material3Dg;
	if(not pMaterial)
	{
		delete[] pTempMaterials;
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	for(uint32_t i = 0; i < mMaterialsQuantity; i++)
		pTempMaterials[i] = mMaterials[i];

	delete[] mMaterials;
	mMaterials = pTempMaterials;
	mMaterials[mMaterialsQuantity] = pMaterial;
	mMaterialsQuantity++;
	return pMaterial;
}

Texture3Dg *I3Dg::createTexture(char *pTexPath, uint16_t uiPathBufferSize)
{
	//new textures array
	Texture3Dg **pTmpTex = new Texture3Dg*[mTexturesQuantity+1];
	if(not pTmpTex)
	{
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	//new texture
	Texture3Dg *pTexture = new Texture3Dg;
	//FIXME: when texture won't be a simple struct it will have to be removed
	memset(pTexture, 0, sizeof(Texture3Dg));
	if(not pTexture)
	{
		delete[] pTmpTex;
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}

	//and the path to it's file
	pTexture->pPath = new char[uiPathBufferSize];
	if(not pTexture->pPath)
	{
		delete pTexture;
		delete[] pTmpTex;
		mError = ERR3DG_OUTOFMEMORY;
		return 0;
	}
	memcpy(pTexture->pPath, pTexPath, uiPathBufferSize * sizeof(char));
	pTexture->uiPathBufferSize = uiPathBufferSize;

	//copy pointers to the new array
	for(uint32_t i = 0; i < mTexturesQuantity; i++)
		pTmpTex[i] = mTextures[i];

	//delete the old array
	delete[] mTextures;
	//and copy the address to the new array into the global pointer
	mTextures = pTmpTex;
	mTextures[mTexturesQuantity] = pTexture;
	mTexturesQuantity++;
	return pTexture;
}

/************************************************\
*                 Material3Dg                    *
\************************************************/

Material3Dg::Material3Dg()
{
	memset(&sDiffuse, 0, sizeof(sDiffuse));
	memset(&sAmbient, 0, sizeof(sAmbient));
	memset(&sSpecular, 0, sizeof(sSpecular));
	memset(&sEmissive, 0, sizeof(sEmissive));
	fShininess = 0.0f;
	uiTexturesQuantity = 0;
	pMtlTextures = 0;
}

Material3Dg::~Material3Dg()
{
	if(pMtlTextures)
		delete[] pMtlTextures;
}

bool Material3Dg::addTexture(int32_t iTexNum)
{
	int32_t *pTmpNums = new int32_t[uiTexturesQuantity + 1];
	if(not pTmpNums)
		return false;

	for(uint32_t i = 0; i < uiTexturesQuantity; i++)
		pTmpNums[i] = pMtlTextures[i];

	pTmpNums[uiTexturesQuantity] = iTexNum;
	delete[] pMtlTextures;
	pMtlTextures = pTmpNums;
	uiTexturesQuantity++;
	return true;
}

int32_t Material3Dg::getTexture(uint32_t uiMtlTexNum)
{
	if(uiMtlTexNum < uiTexturesQuantity)
		return pMtlTextures[uiMtlTexNum];
	else return -1;
}

/************************************************\
*                  Point3Dg                      *
\************************************************/

Point3Dg::Point3Dg()
{
	x = y = z = 0.0f;
}

Point3Dg::Point3Dg(float x, float y, float z)
{
	this->x = x;
	this->y = y;
	this->z = z;
}

/************************************************\
*                  Vector3Dg                     *
\************************************************/

Vector3Dg Vector3Dg::operator +(const Vector3Dg &vec)
{
	return Vector3Dg(x + vec.x, y + vec.y, z  + vec.z);
}

Vector3Dg Vector3Dg::operator -(const Vector3Dg &vec)
{
	return Vector3Dg(x - vec.x, y - vec.y, z  - vec.z);
}

Vector3Dg Vector3Dg::operator *(float multiplier)
{
	return Vector3Dg(x * multiplier, y * multiplier, z * multiplier);
}

float Vector3Dg::dotProduct(const Vector3Dg &vec)
{
	return x * vec.x + y * vec.y + z * vec.z;
}

Vector3Dg Vector3Dg::crossProduct(const Vector3Dg &vec)
{
	return Vector3Dg(y * vec.z - z * vec.y, z * vec.x - x * vec.z, x * vec.y - y * vec.x);
}

float Vector3Dg::length()
{
	return sqrtf(x * x + y * y + z * z);
}

float Vector3Dg::lengthPow2()
{
	return x * x + y * y + z * z;
}

void Vector3Dg::operator +=(Vector3Dg &vec)
{
	x += vec.x;
	y += vec.y;
	z += vec.z;
}

void Vector3Dg::operator -=(Vector3Dg &vec)
{
	x -= vec.x;
	y -= vec.y;
	z -= vec.z;
}

void Vector3Dg::operator *=(float multiplier)
{
	x *= multiplier;
	y *= multiplier;
	z *= multiplier;
}

void Vector3Dg::operator /=(float divisor)
{
	x /= divisor;
	y /= divisor;
	z /= divisor;
}

void Vector3Dg::normalize()
{
	float length = sqrtf((x * x) + (y * y) + (z * z));
	if(length == 0.0f or length == 1.0f)
		return;
	x /= length;
	y /= length;
	z /= length;
}

/************************************************\
*                     Mesh3Dg                    *
\************************************************/

Mesh3Dg::Mesh3Dg(Mesh3DgUID uiUID)
{
	memset(&mTranslation, 0, sizeof(Matrix3Dg));
	mTranslation.R1C1 = 1.0f;
	mTranslation.R2C2 = 1.0f;
	mTranslation.R3C3 = 1.0f;
	mTranslation.R4C4 = 1.0f;
	memset(&mRotation, 0, sizeof(Matrix3Dg));
	mRotation.R1C1 = 1.0f;
	mRotation.R2C2 = 1.0f;
	mRotation.R3C3 = 1.0f;
	mRotation.R4C4 = 1.0f;
	memset(&mScaling, 0, sizeof(Matrix3Dg));
	mScaling.R1C1 = 1.0f;
	mScaling.R2C2 = 1.0f;
	mScaling.R3C3 = 1.0f;
	mScaling.R4C4 = 1.0f;

	memset(&mAnimation, 0, sizeof(MeshAnim3Dg));

	mBase = 0;
	mMaterialNumber = 0;
	mProps = MRP_RENDERABLE;
	mUID = uiUID;
}

Mesh3Dg::~Mesh3Dg()
{}

void Mesh3Dg::rotate(float x, float y, float z)
{
	float fSinX = sinf(x), fCosX = cosf(x);
	float fSinY = sinf(y), fCosY = cosf(y);
	float fSinZ = sinf(z), fCosZ = cosf(z);
	mRotation.R1C1 = fCosY * fCosZ;
	mRotation.R1C2 = fSinZ;
	mRotation.R1C3 = -fSinY;
	mRotation.R2C1 = -fSinZ;
	mRotation.R2C2 = fCosX * fCosZ;
	mRotation.R2C3 = fSinX;
	mRotation.R3C1 = fSinY;
	mRotation.R3C2 = -fSinX;
	mRotation.R3C3 = fCosX * fCosY;
}

void Mesh3Dg::setPosition(float x, float y, float z)
{
	mTranslation.R4C1 = x;
	mTranslation.R4C2 = y;
	mTranslation.R4C3 = z;
}

void Mesh3Dg::addTranslation(float x, float y, float z)
{
	mTranslation.R4C1 += x;
	mTranslation.R4C2 += y;
	mTranslation.R4C3 += z;
}

void Mesh3Dg::scale(float x, float y, float z)
{
	mScaling.R1C1 = x;
	mScaling.R2C2 = y;
	mScaling.R3C3 = z;
}

Vector3Dg Mesh3Dg::geomCenter()
{
	Vector3Dg center(mTranslation.R4C1, mTranslation.R4C2, mTranslation.R4C3);
	if(mProps & MRP_HAS_BASE)
	{
		Triangle3Dg *trians = mBase->getTriangles();
		Vector3Dg basecenter(((trians->sVertexA.sCoords.x + trians->sVertexB.sCoords.x) / 2.0f),
							((trians->sVertexA.sCoords.y + trians->sVertexB.sCoords.y) / 2.0f),
							((trians->sVertexA.sCoords.z + trians->sVertexB.sCoords.z) / 2.0f));
		center += basecenter;
	}
	return center;
}

MeshBase3Dg *Mesh3Dg::setBase(MeshBase3Dg *base)
{
	//previous base
	MeshBase3Dg *prev = mBase;
	//set new base
	mBase = base;
	mProps |= MRP_HAS_BASE;
	//return the previous
	return prev;
}

void Mesh3Dg::setProperty(uint32_t property, bool set)
{
	if(set)
		mProps |= property;
	else
		mProps &= ~property;
}

/************************************************\
*                  MeshBase3Dg                   *
\************************************************/

MeshBase3Dg::MeshBase3Dg(MeshBase3DgUID uid)
{
	mUID = uid;
	mTriansQuantity = 0;
	mTriangles = 0;
}

MeshBase3Dg::~MeshBase3Dg()
{
	releaseTriangles();
}

bool MeshBase3Dg::addTrian(Triangle3Dg *triangle)
{
	Triangle3Dg *pTempTriangles;
	pTempTriangles = new Triangle3Dg[mTriansQuantity + 1];
	if(not pTempTriangles)
		return false;

	for(uint32_t i = 0; i < mTriansQuantity; i++)
		pTempTriangles[i] = mTriangles[i];

	delete[] mTriangles;
	mTriangles = pTempTriangles;

	mTriangles[mTriansQuantity] = *triangle;
	mTriansQuantity++;
	return true;
}

void MeshBase3Dg::releaseTriangles()
{
	if(mTriangles)
		delete[] mTriangles;
	mTriangles = 0;
	mTriansQuantity = 0;
}

Triangle3Dg *MeshBase3Dg::getTriangle(uint32_t trianglenumber)
{
	if(trianglenumber < mTriansQuantity)
		return &mTriangles[trianglenumber];
	else return 0;
}

const char *I3Dg::getLastErrorDesc()
{
	return getErrorDesc(getLastError());
}

const char *I3Dg::getErrorDesc(Error3Dg eError)
{
	switch(eError)
	{
	case ERR3DG_OK:
		return "No error";
	case ERR3DG_OUTOFMEMORY:
		return "Out of memory";
	case ERR3DG_INVALIDCALL:
		return "Invalid call of a function";
	case ERR3DG_INVALIDARG:
		return "Invalid argument to a function";
	default:
		return "Unknown error";
	}
	return 0;
}
